from __future__ import annotations

"""Presentation-phase deterministic helpers."""

from .projectile_decals import queue_projectile_large_streak_decal

__all__ = ["queue_projectile_large_streak_decal"]
