from __future__ import annotations

from enum import IntEnum


class RngCallerStatic(IntEnum):
    BONUS_PICK_RANDOM_TYPE_ROLL = 0x004124A5
    BONUS_PICK_RANDOM_TYPE_ENERGIZER = 0x004124D6
    BONUS_SPAWN_AT_POS_POINTS_AMOUNT = 0x0041F897
    BONUS_APPLY_NUKE_BULLET_COUNT = 0x0040A00B
    BONUS_APPLY_NUKE_PISTOL_ANGLE = 0x0040A023
    BONUS_APPLY_NUKE_PISTOL_SPEED_SCALE = 0x0040A059
    BONUS_APPLY_NUKE_GAUSS_ANGLE_1 = 0x0040A088
    BONUS_APPLY_NUKE_GAUSS_ANGLE_2 = 0x0040A0B4
    BONUS_TRY_SPAWN_ON_KILL_PISTOL_FORCE_WEAPON = 0x0041F92C
    BONUS_TRY_SPAWN_ON_KILL_BASE_GATE = 0x0041F9AA
    BONUS_TRY_SPAWN_ON_KILL_PISTOL_ALLOW_WITHOUT_MAGNET = 0x0041F9C3
    BONUS_TRY_SPAWN_ON_KILL_BONUS_MAGNET = 0x0041F9EB
    BONUS_TRY_SPAWN_ON_KILL_BURST_ROTATION = 0x0041FB0F
    BONUS_TRY_SPAWN_ON_KILL_BURST_VEL_X = 0x0041FB2B
    BONUS_TRY_SPAWN_ON_KILL_BURST_VEL_Y = 0x0041FB4D
    BONUS_TRY_SPAWN_ON_KILL_BURST_SCALE_STEP = 0x0041FB6F

    PLAYER_UPDATE_LOW_HEALTH_BLOODSPILL = 0x0041381C
    PLAYER_UPDATE_CASING_ANGLE = 0x00415A6D
    PLAYER_UPDATE_CASING_SPEED = 0x00415A8B
    PLAYER_UPDATE_CASING_ROTATION = 0x00415B26
    PLAYER_UPDATE_CASING_ROTATION_STEP = 0x00415B79
    PLAYER_UPDATE_SHOT_JITTER_DIR = 0x00415C6B
    PLAYER_UPDATE_SHOT_JITTER_MAG = 0x00415C87
    PLAYER_UPDATE_SHOT_SFX = 0x00415F09
    PLAYER_UPDATE_SHOTGUN_PELLET_JITTER = 0x00416364
    PLAYER_UPDATE_SHOTGUN_PELLET_SPEED_SCALE = 0x00416399
    PLAYER_UPDATE_SAWED_OFF_SHOTGUN_PELLET_JITTER = 0x00416478
    PLAYER_UPDATE_SAWED_OFF_SHOTGUN_PELLET_SPEED_SCALE = 0x004164AD
    PLAYER_UPDATE_JACKHAMMER_PELLET_JITTER = 0x004165FE
    PLAYER_UPDATE_JACKHAMMER_PELLET_SPEED_SCALE = 0x00416633
    PLAYER_UPDATE_ION_SHOTGUN_PELLET_JITTER = 0x00416B44
    PLAYER_UPDATE_ION_SHOTGUN_PELLET_SPEED_SCALE = 0x00416B79
    PLAYER_UPDATE_GAUSS_SHOTGUN_PELLET_JITTER = 0x00416CFD
    PLAYER_UPDATE_GAUSS_SHOTGUN_PELLET_SPEED_SCALE = 0x00416D32
    PLAYER_UPDATE_PLASMA_SHOTGUN_PELLET_JITTER = 0x0041735A
    PLAYER_UPDATE_PLASMA_SHOTGUN_PELLET_SPEED_SCALE = 0x0041738E
    PLAYER_UPDATE_FIRE_COUGH_SPREAD_DIR = 0x00413B16
    PLAYER_UPDATE_FIRE_COUGH_SPREAD_MAG = 0x00413B32
    PLAYER_UPDATE_FIRE_COUGH_INTERVAL_RESET = 0x00413C4F
    PLAYER_UPDATE_HOT_TEMPERED_INTERVAL_RESET = 0x00413D3A
    PLAYER_UPDATE_MAN_BOMB_ION_RIFLE_ANGLE = 0x00413924
    PLAYER_UPDATE_MAN_BOMB_ION_MINIGUN_ANGLE = 0x0041393D

    CREATURE_ALLOC_SLOT_PHASE_SEED = 0x004281A2
    SURVIVAL_SPAWN_CREATURE_TYPE_ROLL = 0x0040754B
    SURVIVAL_SPAWN_CREATURE_PARITY_PICK = 0x004075B8
    SURVIVAL_SPAWN_CREATURE_RARE_OVERRIDE = 0x00407616
    SURVIVAL_SPAWN_CREATURE_SIZE = 0x00407625
    SURVIVAL_SPAWN_CREATURE_HEADING = 0x0040766C
    SURVIVAL_SPAWN_CREATURE_HEALTH = 0x004076D9
    SURVIVAL_SPAWN_CREATURE_LOW_TINT_G = 0x00407792
    SURVIVAL_SPAWN_CREATURE_LOW_TINT_B = 0x004077E3
    SURVIVAL_SPAWN_CREATURE_MID_TINT_G = 0x00407848
    SURVIVAL_SPAWN_CREATURE_MID_TINT_B = 0x00407899
    SURVIVAL_SPAWN_CREATURE_HIGH_TINT_G = 0x00407906
    SURVIVAL_SPAWN_CREATURE_HIGH_TINT_B = 0x00407957
    SURVIVAL_SPAWN_CREATURE_REWARD_BONUS = 0x004079CF
    SURVIVAL_SPAWN_CREATURE_RARE_RED = 0x00407A0A
    SURVIVAL_SPAWN_CREATURE_RARE_GREEN = 0x00407A6C
    SURVIVAL_SPAWN_CREATURE_RARE_BLUE = 0x00407ACB
    SURVIVAL_SPAWN_CREATURE_RARE_PURPLE = 0x00407B28
    SURVIVAL_SPAWN_CREATURE_RARE_YELLOW = 0x00407B93
    SURVIVAL_UPDATE_EXTRA_SPAWN_EDGE = 0x00408338
    SURVIVAL_UPDATE_EXTRA_SPAWN_TOP_X = 0x00408356
    SURVIVAL_UPDATE_EXTRA_SPAWN_BOTTOM_X = 0x00408381
    SURVIVAL_UPDATE_EXTRA_SPAWN_LEFT_Y = 0x004083B2
    SURVIVAL_UPDATE_EXTRA_SPAWN_RIGHT_Y = 0x004083DB
    SURVIVAL_UPDATE_MAIN_SPAWN_EDGE = 0x00408423
    SURVIVAL_UPDATE_MAIN_SPAWN_TOP_X = 0x00408441
    SURVIVAL_UPDATE_MAIN_SPAWN_BOTTOM_X = 0x0040846C
    SURVIVAL_UPDATE_MAIN_SPAWN_LEFT_Y = 0x0040849D
    SURVIVAL_UPDATE_MAIN_SPAWN_RIGHT_Y = 0x004084C6
    CREATURE_SPAWN_HEADING = 0x004282DC
    CREATURE_SPAWN_REWARD = 0x00428315
    CREATURE_HANDLE_DEATH_SPLIT_CHILD_1_PHASE_SEED = 0x0041E9F4
    CREATURE_HANDLE_DEATH_SPLIT_CHILD_2_PHASE_SEED = 0x0041EA91
    CREATURE_HANDLE_DEATH_FREEZE_SHARD_ANGLE = 0x0041EBEA
    CREATURE_HANDLE_DEATH_FREEZE_SHATTER_ANGLE = 0x0041EC15
    CREATURE_SPAWN_TEMPLATE_RANDOM_HEADING = 0x00430B14
    CREATURE_SPAWN_TEMPLATE_BASE_HEADING = 0x00430B88
    CREATURE_SPAWN_TEMPLATE_AI1_BLUE_TINT_1A = 0x004323EC
    CREATURE_SPAWN_TEMPLATE_AI1_BLUE_TINT_1B = 0x00432440
    CREATURE_SPAWN_TEMPLATE_AI1_BLUE_TINT_1C = 0x0043248B
    CREATURE_SPAWN_TEMPLATE_ALIEN_RANDOM_1D_SIZE = 0x0043303B
    CREATURE_SPAWN_TEMPLATE_ALIEN_RANDOM_1D_MOVE_SPEED = 0x00433065
    CREATURE_SPAWN_TEMPLATE_ALIEN_RANDOM_1D_REWARD = 0x00433089
    CREATURE_SPAWN_TEMPLATE_ALIEN_RANDOM_1D_TINT_R = 0x004330AB
    CREATURE_SPAWN_TEMPLATE_ALIEN_RANDOM_1D_TINT_G = 0x004330CF
    CREATURE_SPAWN_TEMPLATE_ALIEN_RANDOM_1D_TINT_B = 0x004330F3
    CREATURE_SPAWN_TEMPLATE_ALIEN_RANDOM_1D_CONTACT_DAMAGE = 0x00433117
    CREATURE_SPAWN_TEMPLATE_ALIEN_RANDOM_1E_SIZE = 0x0043314A
    CREATURE_SPAWN_TEMPLATE_ALIEN_RANDOM_1E_MOVE_SPEED = 0x00433174
    CREATURE_SPAWN_TEMPLATE_ALIEN_RANDOM_1E_REWARD = 0x00433198
    CREATURE_SPAWN_TEMPLATE_ALIEN_RANDOM_1E_TINT_R = 0x004331BA
    CREATURE_SPAWN_TEMPLATE_ALIEN_RANDOM_1E_TINT_G = 0x004331DE
    CREATURE_SPAWN_TEMPLATE_ALIEN_RANDOM_1E_TINT_B = 0x00433202
    CREATURE_SPAWN_TEMPLATE_ALIEN_RANDOM_1E_CONTACT_DAMAGE = 0x00433226
    CREATURE_SPAWN_TEMPLATE_ALIEN_RANDOM_1F_SIZE = 0x00433259
    CREATURE_SPAWN_TEMPLATE_ALIEN_RANDOM_1F_MOVE_SPEED = 0x00433283
    CREATURE_SPAWN_TEMPLATE_ALIEN_RANDOM_1F_REWARD = 0x004332A7
    CREATURE_SPAWN_TEMPLATE_ALIEN_RANDOM_1F_TINT_R = 0x004332C9
    CREATURE_SPAWN_TEMPLATE_ALIEN_RANDOM_1F_TINT_G = 0x004332ED
    CREATURE_SPAWN_TEMPLATE_ALIEN_RANDOM_1F_TINT_B = 0x00433311
    CREATURE_SPAWN_TEMPLATE_ALIEN_RANDOM_1F_CONTACT_DAMAGE = 0x00433335
    CREATURE_SPAWN_TEMPLATE_ALIEN_RANDOM_GREEN_20_SIZE = 0x00432857
    CREATURE_SPAWN_TEMPLATE_ALIEN_RANDOM_GREEN_20_MOVE_SPEED = 0x00432881
    CREATURE_SPAWN_TEMPLATE_ALIEN_RANDOM_GREEN_20_TINT_G = 0x004328C2
    CREATURE_SPAWN_TEMPLATE_ALIEN_RANDOM_GREEN_20_CONTACT_DAMAGE = 0x004328E9
    CREATURE_SPAWN_TEMPLATE_SPIDER_SP1_RANDOM_03_SIZE = 0x0043291C
    CREATURE_SPAWN_TEMPLATE_SPIDER_SP1_RANDOM_03_MOVE_SPEED = 0x00432946
    CREATURE_SPAWN_TEMPLATE_SPIDER_SP1_RANDOM_03_TINT_B = 0x0043298B
    CREATURE_SPAWN_TEMPLATE_SPIDER_SP1_RANDOM_03_CONTACT_DAMAGE = 0x00432A4F
    CREATURE_SPAWN_TEMPLATE_SPIDER_SP2_RANDOM_05_SIZE = 0x00432A82
    CREATURE_SPAWN_TEMPLATE_SPIDER_SP2_RANDOM_05_MOVE_SPEED = 0x00432AAC
    CREATURE_SPAWN_TEMPLATE_SPIDER_SP2_RANDOM_05_TINT_B = 0x00432AF1
    CREATURE_SPAWN_TEMPLATE_SPIDER_SP2_RANDOM_05_CONTACT_DAMAGE = 0x00432BB5
    CREATURE_SPAWN_TEMPLATE_LIZARD_RANDOM_04_SIZE = 0x00432BE8
    CREATURE_SPAWN_TEMPLATE_LIZARD_RANDOM_04_MOVE_SPEED = 0x00432C12
    CREATURE_SPAWN_TEMPLATE_LIZARD_RANDOM_04_CONTACT_DAMAGE = 0x00432C5A
    CREATURE_SPAWN_TEMPLATE_ALIEN_RANDOM_06_SIZE = 0x00432C8D
    CREATURE_SPAWN_TEMPLATE_ALIEN_RANDOM_06_MOVE_SPEED = 0x00432CB7
    CREATURE_SPAWN_TEMPLATE_ALIEN_RANDOM_06_TINT_B = 0x00432CFC
    CREATURE_SPAWN_TEMPLATE_ALIEN_RANDOM_06_CONTACT_DAMAGE = 0x00432DC0
    CREATURE_SPAWN_TEMPLATE_LIZARD_RANDOM_31_SIZE = 0x00432579
    CREATURE_SPAWN_TEMPLATE_LIZARD_RANDOM_31_MOVE_SPEED = 0x004325A3
    CREATURE_SPAWN_TEMPLATE_LIZARD_RANDOM_31_TINT = 0x004325DC
    CREATURE_SPAWN_TEMPLATE_SPIDER_SP1_RANDOM_32_SIZE = 0x0043261F
    CREATURE_SPAWN_TEMPLATE_SPIDER_SP1_RANDOM_32_MOVE_SPEED = 0x00432643
    CREATURE_SPAWN_TEMPLATE_SPIDER_SP1_RANDOM_32_TINT = 0x0043267C
    CREATURE_SPAWN_TEMPLATE_SPIDER_SP1_RANDOM_RED_33_SIZE = 0x004326CD
    CREATURE_SPAWN_TEMPLATE_SPIDER_SP1_RANDOM_RED_33_MOVE_SPEED = 0x004326F7
    CREATURE_SPAWN_TEMPLATE_SPIDER_SP1_RANDOM_RED_33_TINT_R = 0x00432730
    CREATURE_SPAWN_TEMPLATE_SPIDER_SP1_RANDOM_RED_33_CONTACT_DAMAGE = 0x0043275F
    CREATURE_SPAWN_TEMPLATE_SPIDER_SP1_RANDOM_GREEN_34_SIZE = 0x00432792
    CREATURE_SPAWN_TEMPLATE_SPIDER_SP1_RANDOM_GREEN_34_MOVE_SPEED = 0x004327BC
    CREATURE_SPAWN_TEMPLATE_SPIDER_SP1_RANDOM_GREEN_34_TINT_G = 0x004327F5
    CREATURE_SPAWN_TEMPLATE_SPIDER_SP1_RANDOM_GREEN_34_CONTACT_DAMAGE = 0x00432824
    CREATURE_SPAWN_TEMPLATE_LIZARD_RANDOM_2E_SIZE = 0x00432EB8
    CREATURE_SPAWN_TEMPLATE_LIZARD_RANDOM_2E_MOVE_SPEED = 0x00432EE2
    CREATURE_SPAWN_TEMPLATE_LIZARD_RANDOM_2E_TINT_R = 0x00432F1B
    CREATURE_SPAWN_TEMPLATE_LIZARD_RANDOM_2E_TINT_G = 0x00432F3F
    CREATURE_SPAWN_TEMPLATE_LIZARD_RANDOM_2E_TINT_B = 0x00432F63
    CREATURE_SPAWN_TEMPLATE_LIZARD_RANDOM_2E_CONTACT_DAMAGE = 0x00432F87
    CREATURE_SPAWN_TEMPLATE_SPIDER_SP2_RANDOM_35_SIZE = 0x00432DF3
    CREATURE_SPAWN_TEMPLATE_SPIDER_SP2_RANDOM_35_MOVE_SPEED = 0x00432E1D
    CREATURE_SPAWN_TEMPLATE_SPIDER_SP2_RANDOM_35_TINT_G = 0x00432E5E
    CREATURE_SPAWN_TEMPLATE_SPIDER_SP2_RANDOM_35_CONTACT_DAMAGE = 0x00432E85
    CREATURE_SPAWN_TEMPLATE_AI7_ORBITER_TINT_G = 0x00432FED
    CREATURE_SPAWN_TEMPLATE_SPIDER_SP1_RANDOM_3D_TINT = 0x00433B39
    CREATURE_SPAWN_TEMPLATE_SPIDER_SP1_RANDOM_3D_SIZE = 0x00433B63
    CREATURE_SPAWN_TEMPLATE_SPIDER_SP1_AI7_TIMER_38_SIZE = 0x00433D18
    CREATURE_SPAWN_TEMPLATE_SPIDER_SP2_RANGED_VARIANT_37_SIZE = 0x00433DB3
    CREATURE_SPAWN_TEMPLATE_SPIDER_SP1_AI7_TIMER_WEAK_39_SIZE = 0x00433E4E
    CREATURE_SPAWN_TEMPLATE_ZOMBIE_RANDOM_41_SIZE = 0x004324CD
    CREATURE_SPAWN_TEMPLATE_ZOMBIE_RANDOM_41_TINT = 0x0043251C
    CREATURE_SPAWN_TEMPLATE_ZOMBIE_RANDOM_41_CONTACT_DAMAGE = 0x00432546

    CREATURE_UPDATE_ALL_AI7_LINK_TIMER_HOLD = 0x00426381
    CREATURE_UPDATE_ALL_AI7_LINK_TIMER_RESET = 0x004263B1
    CREATURE_UPDATE_ALL_CONTACT_SFX = 0x00426611
    CREATURE_UPDATE_ALL_PLASMAMINIGUN_COOLDOWN = 0x00427119
    CREATURE_UPDATE_ALL_PLAGUE_KILL_SFX = 0x00427273
    CREATURE_UPDATE_ALL_PING_PONG_BLOOD_8_ANGLE = 0x004275A3
    CREATURE_UPDATE_ALL_PING_PONG_BLOOD_6_ANGLE = 0x004275D8
    CREATURE_UPDATE_ALL_PING_PONG_BLOOD_5_ANGLE = 0x0042760D

    CREATURE_APPLY_DAMAGE_HEADING_JITTER = 0x004208A9
    CREATURE_APPLY_DAMAGE_PYROMANIAC = 0x0042092D
    CREATURE_APPLY_DAMAGE_DEATH_SFX = 0x00420B16
    CREATURE_APPLY_DAMAGE_SHOCK_BURST_ROTATION = 0x00420A4B
    CREATURE_APPLY_DAMAGE_SHOCK_BURST_VEL_X = 0x00420A67
    CREATURE_APPLY_DAMAGE_SHOCK_BURST_VEL_Y = 0x00420A80
    CREATURE_APPLY_DAMAGE_SHOCK_BURST_SCALE_STEP = 0x00420A99

    BONUS_APPLY_FREEZE_SHARD_ANGLE = 0x00409BDC
    BONUS_APPLY_FREEZE_SHATTER_ANGLE = 0x00409C07

    PROJECTILE_UPDATE_POISON_BULLETS_GATE = 0x00420E84
    PROJECTILE_UPDATE_BLADE_GUN_SPLATTER_ANGLE = 0x00420EAF
    PROJECTILE_UPDATE_BLOODY_MESS_SPREAD = 0x00420FD7
    PROJECTILE_UPDATE_DEFAULT_REVERSE_SPLATTER_GATE = 0x0042105B
    PROJECTILE_UPDATE_BLOODY_MESS_DECAL_DX_1 = 0x004210D3
    PROJECTILE_UPDATE_BLOODY_MESS_DECAL_DY_1 = 0x004210EE
    PROJECTILE_UPDATE_BLOODY_MESS_DECAL_DX_2 = 0x00421130
    PROJECTILE_UPDATE_BLOODY_MESS_DECAL_DY_2 = 0x0042114B
    PROJECTILE_UPDATE_STOP_ON_HIT_JITTER = 0x004211C4
    PROJECTILE_UPDATE_POST_HIT_DECAL_BURN = 0x004215D2
    PROJECTILE_UPDATE_DEFAULT_FREEZE_SHARD_ANGLE = 0x004215FF
    PROJECTILE_UPDATE_DECAL_SPREAD = 0x00421642
    PROJECTILE_UPDATE_LARGE_STREAK_DIST = 0x0042176F
    PROJECTILE_UPDATE_LARGE_STREAK_DIST_GT4 = 0x00421799
    PROJECTILE_UPDATE_LARGE_STREAK_DIST_GT7 = 0x004217C6
    PROJECTILE_UPDATE_LARGE_STREAK_BURN = 0x0042184C
    PROJECTILE_UPDATE_LARGE_STREAK_FREEZE_ANGLE = 0x00421879
    PROJECTILE_UPDATE_HIT_SFX = 0x00421936
    PROJECTILE_UPDATE_PARTICLE_JITTER_FLAMETHROWER = 0x0042277A
    PROJECTILE_UPDATE_PARTICLE_JITTER_BUBBLEGUN = 0x004227AE
    PROJECTILE_UPDATE_PARTICLE_JITTER_ALT = 0x004227F9
    PROJECTILE_UPDATE_PARTICLE_BOUNCE_SPEED_SCALE = 0x00422A01
    PROJECTILE_UPDATE_PARTICLE_SPRITE_VEL_X = 0x00422BA8
    PROJECTILE_UPDATE_PARTICLE_SPRITE_VEL_Y = 0x00422BBC
    SECONDARY_PROJECTILE_UPDATE_PRE_HIT_DECAL_DX_1 = 0x00421EEB
    SECONDARY_PROJECTILE_UPDATE_PRE_HIT_DECAL_DY_1 = 0x00421F0A
    SECONDARY_PROJECTILE_UPDATE_PRE_HIT_DECAL_DX_2 = 0x00421F47
    SECONDARY_PROJECTILE_UPDATE_PRE_HIT_DECAL_DY_2 = 0x00421F66
    SECONDARY_PROJECTILE_UPDATE_PRE_HIT_DECAL_DX_3 = 0x00421FAC
    SECONDARY_PROJECTILE_UPDATE_PRE_HIT_DECAL_DY_3 = 0x00421FCB
    SECONDARY_PROJECTILE_UPDATE_PRE_HIT_FREEZE_SHARD_ANGLE = 0x00421EB3
    SECONDARY_PROJECTILE_UPDATE_ROCKET_FREEZE_SHARD_ANGLE = 0x00422130
    SECONDARY_PROJECTILE_UPDATE_ROCKET_DECAL_ANGLE = 0x00422165
    SECONDARY_PROJECTILE_UPDATE_ROCKET_DECAL_RADIUS = 0x00422184
    SECONDARY_PROJECTILE_UPDATE_SEEKER_ROCKET_FREEZE_SHARD_ANGLE = 0x0042221E
    SECONDARY_PROJECTILE_UPDATE_SEEKER_ROCKET_DECAL_ANGLE = 0x00422253
    SECONDARY_PROJECTILE_UPDATE_SEEKER_ROCKET_DECAL_RADIUS = 0x00422272
    SECONDARY_PROJECTILE_UPDATE_ROCKET_MINIGUN_FREEZE_SHARD_ANGLE = 0x0042230D
    SECONDARY_PROJECTILE_UPDATE_ROCKET_MINIGUN_DECAL_ANGLE = 0x0042233F
    SECONDARY_PROJECTILE_UPDATE_ROCKET_MINIGUN_DECAL_RADIUS = 0x0042235E
    SECONDARY_PROJECTILE_UPDATE_DETONATION_SPRITE_MAG = 0x004223BE

    UNLOCK_TERRAIN_Q4 = 0x00418229
    UNLOCK_TERRAIN_Q3 = 0x00418254
    UNLOCK_TERRAIN_Q2 = 0x0041827F

    PLAYER_TAKE_DAMAGE_NINJA = 0x00425F23
    PLAYER_TAKE_DAMAGE_DODGER = 0x00425F4E
    PLAYER_TAKE_DAMAGE_HIGHLANDER = 0x00425F79
    PLAYER_TAKE_DAMAGE_DEATH_SFX = 0x00426124
    PLAYER_TAKE_DAMAGE_PAIN_SFX = 0x00426152
    PLAYER_TAKE_DAMAGE_HEADING = 0x00426197
    PLAYER_TAKE_DAMAGE_LOW_HEALTH = 0x00426201

    FX_QUEUE_ADD_RANDOM_GRAY = 0x00427760
    FX_QUEUE_ADD_RANDOM_WIDTH = 0x0042778E
    FX_QUEUE_ADD_RANDOM_ROTATION = 0x004277B0
    FX_QUEUE_ADD_RANDOM_EFFECT_ID = 0x0042780B

    FX_SPAWN_PARTICLE_ALLOC = 0x00420151
    FX_SPAWN_PARTICLE_SPIN = 0x00420206
    FX_SPAWN_PARTICLE_SLOW_ALLOC = 0x00420261
    FX_SPAWN_PARTICLE_SLOW_SPIN = 0x00420316
    FX_SPAWN_SPRITE_ALLOC = 0x0041FBCE
    FX_SPAWN_SPRITE_ROTATION = 0x0041FC3E

    EFFECT_SPAWN_BLOOD_SPLATTER_ROTATION = 0x0042EBC0
    EFFECT_SPAWN_BLOOD_SPLATTER_HALF = 0x0042EBE3
    EFFECT_SPAWN_BLOOD_SPLATTER_SPEED_X = 0x0042EC00
    EFFECT_SPAWN_BLOOD_SPLATTER_SPEED_Y = 0x0042EC1D
    EFFECT_SPAWN_BLOOD_SPLATTER_SCALE_STEP = 0x0042EC44

    EFFECT_SPAWN_BURST_ROTATION = 0x0042EFF2
    EFFECT_SPAWN_BURST_VEL_X = 0x0042F00E
    EFFECT_SPAWN_BURST_VEL_Y = 0x0042F027
    EFFECT_SPAWN_BURST_SCALE_STEP = 0x0042F040

    EFFECT_SPAWN_FREEZE_SHARD_LIFETIME = 0x0042ECD8
    EFFECT_SPAWN_FREEZE_SHARD_ROTATION = 0x0042ED26
    EFFECT_SPAWN_FREEZE_SHARD_HALF = 0x0042ED4B
    EFFECT_SPAWN_FREEZE_SHARD_ROTATION_STEP = 0x0042ED93
    EFFECT_SPAWN_FREEZE_SHARD_SCALE_STEP = 0x0042EDC0
    EFFECT_SPAWN_FREEZE_SHARD_EFFECT_ID = 0x0042EDE3

    EFFECT_SPAWN_FREEZE_SHATTER_HALF = 0x0042EEB8
    EFFECT_SPAWN_FREEZE_SHATTER_ROTATION_STEP = 0x0042EEDC
    EFFECT_SPAWN_FREEZE_SHATTER_SHARD_ANGLE = 0x0042EF27

    EFFECT_SPAWN_EXPLOSION_BURST_PUFF_ROTATION = 0x0042F848
    EFFECT_SPAWN_EXPLOSION_BURST_ROTATION = 0x0042F9E6
    EFFECT_SPAWN_EXPLOSION_BURST_VEL_X = 0x0042FA07
    EFFECT_SPAWN_EXPLOSION_BURST_VEL_Y = 0x0042FA21
    EFFECT_SPAWN_EXPLOSION_BURST_SCALE_STEP = 0x0042FA3B
    EFFECT_SPAWN_EXPLOSION_BURST_ROTATION_STEP = 0x0042FA58

    PERK_SELECT_RANDOM = 0x0042FBDC
    PERKS_GENERATE_CHOICES_RARITY_GATE = 0x004046D4
    PERK_APPLY_FATAL_LOTTERY = 0x0040562F
    PERK_APPLY_BANDAGE_HEAL = 0x004058C9
    PERKS_UPDATE_EFFECTS_REGENERATION_GATE = 0x00406B62
    PERKS_UPDATE_EFFECTS_PYROKINETIC_ANGLE_0P8 = 0x00406DFF
    PERKS_UPDATE_EFFECTS_PYROKINETIC_ANGLE_0P6 = 0x00406E41
    PERKS_UPDATE_EFFECTS_PYROKINETIC_ANGLE_0P4 = 0x00406E83
    PERKS_UPDATE_EFFECTS_PYROKINETIC_ANGLE_0P3 = 0x00406EC5
    PERKS_UPDATE_EFFECTS_PYROKINETIC_ANGLE_0P2 = 0x00406F07
    PERKS_UPDATE_EFFECTS_JINXED_ACCIDENT_GATE = 0x00406FB4
    PERKS_UPDATE_EFFECTS_JINXED_TIMER_RESET = 0x00406FEF
    PERKS_UPDATE_EFFECTS_JINXED_CREATURE_PICK = 0x00407033
    PERKS_UPDATE_EFFECTS_JINXED_CREATURE_RETRY = 0x00407055
    QUEST_START_SELECTED_HIGHSCORE_RANDOM_TAG = 0x0043A81B
    QUEST_BUILD_TARGET_PRACTICE_ANGLE = 0x00437A19
    QUEST_BUILD_TARGET_PRACTICE_RADIUS = 0x00437A38
    QUEST_BUILD_THE_RANDOM_FACTOR_BRUTE_GATE = 0x004363DF
    QUEST_BUILD_SWEEP_STAKES_ANGLE = 0x00437838
    QUEST_BUILD_DEJA_VU_ANGLE = 0x00437946
    QUEST_BUILD_THE_KILLING_TEMPLATE_PICK = 0x004384BE
    QUEST_BUILD_THE_KILLING_LAYOUT_PICK = 0x004384F0
    QUEST_BUILD_THE_KILLING_SPAWNER_1_Y = 0x004385D7
    QUEST_BUILD_THE_KILLING_SPAWNER_1_X = 0x004385F5
    QUEST_BUILD_THE_KILLING_SPAWNER_2_Y = 0x00438628
    QUEST_BUILD_THE_KILLING_SPAWNER_2_X = 0x00438646
    QUEST_BUILD_THE_KILLING_SPAWNER_3_Y = 0x0043867F
    QUEST_BUILD_THE_KILLING_SPAWNER_3_X = 0x0043869D

    TYPO_WORD_PICK_FRAGMENT = 0x00444F88
    TYPO_WORD_PICK_HIGHSCORE_NAME = 0x004452EB
    TYPO_TARGET_NAME_ASSIGN_RANDOM_HIGHSCORE_GATE = 0x004453A8
    TYPO_TARGET_NAME_ASSIGN_RANDOM_FOUR_WORD_GATE = 0x004453F6
    TYPO_TARGET_NAME_ASSIGN_RANDOM_THREE_WORD_GATE_GT80 = 0x0044545B
    TYPO_TARGET_NAME_ASSIGN_RANDOM_THREE_WORD_GATE_GT60 = 0x004454B5
    TYPO_TARGET_NAME_ASSIGN_RANDOM_TWO_WORD_GATE_GT40 = 0x004454D0
    TYPO_TARGET_NAME_ASSIGN_RANDOM_TWO_WORD_GATE_GT20 = 0x0044551C
    TYPO_GAMEPLAY_TYPECLICK_BACKSPACE = 0x0044593F
    TYPO_GAMEPLAY_TYPECLICK_CHAR = 0x004459A1
    CREATURE_SPAWN_TINTED_HEADING = 0x004448A3
    CREATURE_SPAWN_TINTED_SIZE = 0x004448FE
    UI_TEXT_INPUT_UPDATE_TYPECLICK = 0x0043EDC2
    CREDITS_SECRET_ALIEN_ZOOKEEPER_FILL_EMPTY = 0x0040FD61
    CREDITS_SECRET_ALIEN_ZOOKEEPER_REROLL_FILL = 0x0040FE2C
    SFX_PLAY_EXCLUSIVE_PLAYLIST_PICK = 0x0043D4BD

    CAMERA_UPDATE_OFFSET_X_BASE = 0x00409578
    CAMERA_UPDATE_OFFSET_X_SPREAD = 0x004095A7
    CAMERA_UPDATE_OFFSET_X_SIGN = 0x004095BA
    CAMERA_UPDATE_OFFSET_Y_BASE = 0x004095D9
    CAMERA_UPDATE_OFFSET_Y_SPREAD = 0x00409608
    CAMERA_UPDATE_OFFSET_Y_SIGN = 0x0040961B
    DEMO_SETUP_VARIANT_1_SPIDER_SP1_X = 0x0040311C
    DEMO_SETUP_VARIANT_1_SPIDER_SP1_Y = 0x00403138
    DEMO_SETUP_VARIANT_1_SPIDER_SP2_X = 0x0040317E
    DEMO_SETUP_VARIANT_1_SPIDER_SP2_Y = 0x0040319A
    DEMO_SETUP_VARIANT_3_ALIEN_BIG_X = 0x0040327C
    DEMO_SETUP_VARIANT_3_ALIEN_BIG_Y = 0x00403298
    DEMO_SETUP_VARIANT_3_ALIEN_SMALL_X = 0x004032DB
    DEMO_SETUP_VARIANT_3_ALIEN_SMALL_Y = 0x004032F7

    SHRINKIFIER_HIT_ROTATION = 0x0042F1CE
    SHRINKIFIER_HIT_VEL_X = 0x0042F1EA
    SHRINKIFIER_HIT_VEL_Y = 0x0042F209
    SHRINKIFIER_HIT_SCALE_STEP = 0x0042F228

    SPLITTER_HIT_ANGLE = 0x0042F4A2
    SPLITTER_HIT_RADIUS = 0x0042F4C4
    SPLITTER_HIT_AGE = 0x0042F4F3

    ION_HIT_SPARK_ROTATION = 0x0042F61A
    ION_HIT_SPARK_VEL_X = 0x0042F636
    ION_HIT_SPARK_VEL_Y = 0x0042F659
    ION_HIT_SPARK_SCALE_STEP = 0x0042F67C

    WEAPON_PICK_RANDOM_AVAILABLE_PICK = 0x00452CD6
    WEAPON_PICK_RANDOM_AVAILABLE_REROLL_GATE = 0x00452CF1
    WEAPON_PICK_RANDOM_AVAILABLE_REROLL_PICK = 0x00452CFA

    # Synthetic caller ids reserve the high range so they cannot be
    # mistaken for native return addresses from the main executable.
    REWRITE_JINXED_ACCIDENT_TARGET_PICK = 0xF1000001
    REPLAY_DRIVER_INTER_TICK_DRAW_BY_TICK = 0xF1000004
    REPLAY_DRIVER_INTER_TICK_DRAW = 0xF1000005
    REWRITE_STATS_MENU_EASTER_ROLL = 0xF1000006
    REWRITE_STATS_MENU_EASTER_TEXT_X = 0xF1000007


__all__ = ["RngCallerStatic"]
